from __future__ import print_function
from maestro.core import execute
from maestro.aws import s3

import sys,os

#Internal Modules
from . import fetch, unpack, deploy

#Config
from . import config

#Entry point for Umpire
def entry():
    exit_code = Umpire().exit_code
    sys.exit(exit_code)

class Umpire(execute.ModuleExecuter):
    
    #
    # !!!Register all the dependencies of this program here.!!!
    #
    def register_dependencies(self):
        self.register(s3.AsyncS3Downloader())
        self.register(fetch.FetchModule())
        self.register(unpack.UnpackModule())
        self.register(deploy.DeploymentModule())

    def run(self, kwargs):
        self.register_dependencies()

        #Get the instance of our first module
        deployer = self.getObject("deploy")

        #TODO: Fix to use just RUN
        deployer.cache_root = self.get_cache_root()

        #Run it
        self.exit_code = deployer.run(kwargs)


    def get_cache_root(self):
        hardcoded_root = "./umpirecache_tmp"
        try:
            if not os.path.exists(config.default_cache_location):
                os.makedirs(config.default_cache_location)
            return config.default_cache_location
        except Exception as e:
            print("Umpire: Error creating default cache location, using local directory './umpirecache_tmp': " + str(e), file=sys.stderr)
            if not os.path.exists(hardcoded_root):
                os.makedirs(hardcoded_root)
            return hardcoded_root




"""deploy.py"""

HELPTEXT = """
                  ----- Umpire -----

The deploy module reads an umpire deployment JSON file.

Usage: umpire <deployment_file>

"""

import sys, os, json
from maestro.core import module
from maestro.tools import path

#Local modules
import fetch, cache
from cache import CacheError

HELP_KEYS = ["h", "help"]

class DeploymentError(Exception):
    pass

class DeploymentModule(module.AsyncModule):
    # Required ID of this module
    id = "deploy"

    #Cache Root
    cache_root = None

    #Set to true to view tracebacks for exceptions
    DEBUG = False

    def help(self):
        print(self.help_text)
        exit(0)


    def run(self,kwargs):
        json_index = 1
        try:
            for index, item in enumerate(sys.argv):
                if index == 0:
                    continue
                if item == "-d" or item == "--debug":
                    self.DEBUG = True
                else:
                    json_index = index

            with open(sys.argv[json_index]) as f:
                data = json.load(f)
        except IndexError:
            print(HELPTEXT)
            sys.exit(1)
        except IOError as e:
            if not self.DEBUG:
                print("Unable to locate file: " + sys.argv[json_index])
            else:
                raise e
            sys.exit(1)

        fetchers = list()

        for repo in data:
            repo_url = repo["url"]
            for item in repo["items"]:
                name = item["name"]
                version = item["version"]
                platform = item["platform"]
                destination = item["destination"]

                fetcher = fetch.FetchModule(None)
                fetcher.dependency_name = name
                fetcher.dependency_version = version
                fetcher.dependency_platform = platform
                fetcher.dependency_repo = repo_url
                fetcher.cache_root = self.cache_root

                #TODO: Figure out how to move this out of deploy
                try:
                    cache_dir =  os.path.join(fetcher.cache_root, fetcher.get_cache_name())
                    cache.create_local_cache(cache_dir, repo_url)
                except CacheError:
                    pass #Cache already exists
                fetcher.start()
                fetchers.append((fetcher,destination))

        done_count = 0
        exit_code = 0
        while done_count < len(fetchers):
            done_count = 0
            for fetcher, destination in fetchers:
                if not os.path.exists(destination):
                    os.makedirs(destination)
                if fetcher.exception is not None:
                    if self.DEBUG:
                        print (fetcher.exception.traceback)
                    else:
                        print (fetcher.format_entry_name() + ": ERROR -- " + str(fetcher.exception))
                    exit_code = 1
                    fetcher.status = module.PROCESSED
                if fetcher.status == module.DONE and fetcher.exception is None:
                    #Check for an exception, raise the full trace if in debug
                    for entry in fetcher.result:
                        destination_file = os.path.join(destination,os.path.split(entry)[1])
                        if os.path.exists(destination_file) or os.path.islink(destination_file):
                            print (fetcher.format_entry_name() + ": Already deployed.")
                            fetcher.status = module.PROCESSED
                            break
                        print (fetcher.format_entry_name() + ": Linking " + destination_file)
                        path.symlink(entry, destination_file)

                        #TODO: Kinda hacky, no significance other than to make it not DONE
                        fetcher.status = module.PROCESSED

                if fetcher.status == module.PROCESSED:
                    done_count += 1
        return exit_code



"""fetch.py"""

import sys, os, urllib, time, traceback
from cache import LocalCache
from multiprocessing import Value
from maestro.core import module
from maestro.aws import s3

CACHE_LOCATION_KEYS = ["c", "cache-location"]
DEPENDENCY_NAME_KEYS = ["n", "name"]
DEPENDENCY_PLATFORM_KEYS = ["p", "platform"]
DEPENDENCY_REPOSITORY_KEYS = ["r", "repository"]
DEPENDENCY_VERSION_KEYS = ["v", "version"]
HELP_KEYS = ["h", "help"]

class EntryError(Exception):
    pass

class FetchModule(module.AsyncModule):

    HELPTEXT = """
                  ----- Fetch Module -----

The fetch module locates a "package" using arbitrary identifiers. It will return a local directory.

-c, --cache-root <path>:        Specify the local root cache directory (must exist)
                                    (required)
-n, --name <name>:              Specify the dependency name
                                    (default source file name)
-p, --platform <platform>:      Specify the target platform
                                    (required)
-r, --repository <url>:          Specify the root of the repository to obtain files.
                                    (required)
-v, --version <version>:        Specify the target version
                                    (default latest)

"""

    # Required ID of this module
    id = "fetch"

    #Initialize shm value #TODO: Near future
    progress = Value('d', 0.0)

    # Version of dependency (Default latest) #TODO: Future
    dependency_version = None

    ### Required Variables for this Module ###
    #Location of the cache root folder (r+w by every user)
    cache_root = None

    #Name of dependency to fetch
    dependency_name = None

    #Name of platform of dependency to fetch
    dependency_platform = None

    #The repository to fetch the dependency, or to verify the integrity from
    dependency_repo = None

    def run(self,kwargs):
        if kwargs is not None:
            pass #TODO: Near future

        #Verify argument validity
        self.__verify_arguments__()

        #Get cache name from remote url
        cache_name = self.get_cache_name()

        #Get cache path
        cache_path = os.path.join(self.cache_root, cache_name)

        #Get cache object (will raise an exception if it doesn't exist)
        cache = LocalCache(cache_path)

        #Try to get entry from cache
        entry = cache.get(self.dependency_platform, self.dependency_name, self.dependency_version)

        if entry is None:
            print (self.format_entry_name() + ": Not in cache")
            full_url = s3.join_s3_url(self.dependency_repo, self.dependency_platform, self.dependency_name, self.dependency_version)

            print (self.format_entry_name() + ": Downloading " + full_url)

            #Get Downloader
            downloader = s3.AsyncS3Downloader(None)

            #Set Downloader arguments
            downloader.source_url = full_url
            downloader.destination_path = os.path.join(self.cache_root, "downloading") + os.sep
            downloader.start()

            #Wait for downloader to finish #TODO: Do something with the reported progress
            while downloader.status != module.DONE:
                time.sleep(0.5)

            #Check for an exception, if so bubble it up
            if downloader.exception is not None:
                raise downloader.exception

            print self.format_entry_name() + ": Download complete"

            if downloader.result is None or len(downloader.result) == 0:
                raise EntryError(self.format_entry_name() + ": Unable to find remote entry '" + full_url + "'")

            #Iterate of the result (downloaded files)
            for item in downloader.result:
                #TODO: MD5 verification
                print (self.format_entry_name() + ": Unpacking...")
                cache.put(item,self.dependency_platform, self.dependency_name, self.dependency_version)
            entry = cache.get(self.dependency_platform, self.dependency_name, self.dependency_version)
            if entry is None:
                raise EntryError(self.format_entry_name() + ": Error retrieving entry from cache.")
        #Entry is not None, return all the files listed in the entry that aren't the configuration files
        return [os.path.abspath(os.path.join(entry.path,f)) for f in os.listdir(entry.path) if f != ".umpire"]


    def format_entry_name(self):
        return str(self.dependency_platform) + "/" + str(self.dependency_name) + " v" + str(self.dependency_version)

    def __verify_arguments__(self):
        if not self.dependency_repo:
            raise ValueError("You must specify a valid repository URL.")
        if not self.dependency_name:
            raise ValueError("You must specify a valid name for the dependency.")
        if not self.dependency_platform:
            raise ValueError("You must specify a valid platform for the dependency.")
        if not self.dependency_version:
            raise ValueError("You must specify a valid version for the dependency.")

    def get_cache_name(self):
        try:
            bucket, prefix = s3.parse_s3_url(self.dependency_repo)
            return bucket + ".s3"
        except ValueError:
            raise ValueError("Only URLs formatted as: s3://{bucket}/ are currently accepted.")

    def __parse_kwargs__(self,kwargs):
        if kwargs is None:
            return
        for key, val in kwargs.iteritems():
            if key in CACHE_LOCATION_KEYS:
                self.cache_repo = val
            elif key in DEPENDENCY_NAME_KEYS:
                self.dependency_name = val
            elif key in DEPENDENCY_PLATFORM_KEYS:
                self.dependency_platform = val
            elif key in DEPENDENCY_REPOSITORY_KEYS:
                self.dependency_repo = val
            elif key in DEPENDENCY_VERSION_KEYS:
                self.dependency_version = val

if __name__ == "__main__":
    from cache import create_local_cache

    remote_url = "s3://thirdpartydependencies/"
    create_local_cache("./test_cache/thirdpartydependencies.s3", remote_url)
    fetcher = FetchModule(None)
    fetcher.dependency_repo = remote_url
    fetcher.cache_root = "./test_cache"
    fetcher.dependency_name = "zlib"
    fetcher.dependency_version = "1.2.8"
    fetcher.dependency_platform = "win64"

    fetcher.start()
    while fetcher.status != module.DONE:
        time.sleep(0.5)

    if fetcher.exception is not None:
        print (str(fetcher.exception))
    for item in fetcher.result:
        print (item)
    print ("Done!")



"""unpack.py"""

HELPTEXT = """
     ----- Unpack Module -----

The unpack module unpacks a tar gz file.

Usage: Currently N/A

"""

import sys, os
from maestro.core import module

HELP_KEYS = ["h", "help"]

class UnpackError(Exception):
    pass

class UnpackModule(module.AsyncModule):
    # Required ID of this module
    id = "unpack"
    file_path = None
    destination_path = None
    delete_archive = False

    def help(self):
        print(self.help_text)
        exit(0)

    def run(self,kwargs):
        if self.file_path.endswith("tar.gz") or self.file_path.endswith("tgz"):
            import tarfile
            with tarfile.open(self.file_path, "r:gz") as f:
                f.extractall(self.destination_path)
            if self.delete_archive is True:
                os.remove(self.file_path)
        else:
            raise UnpackError("Unable to unpack this type of file: " + os.path.split(self.file_path)[1])



"""
repo.py contains code to control the local cache for umpire. It is not a module.
"""
#TODO: Add support for zip, tbz/tar.bz 
#TODO: Write command line module

import os, shutil, time
import ConfigParser
import maestro.tools.path
import maestro.core.module
from urlparse import urlparse
from unpack import UnpackModule

CONFIG_FILENAME = ".umpire"
CONFIG_REPO_SECTION_NAME = "umpire"
CONFIG_ENTRY_SECTION_NAME = "entry"

CURRENT_ENTRY_CONFIG_VERSION = "0.1"
CURRENT_REPO_CONFIG_VERSION = "0.1"

def create_local_cache(local_path, remote_url):
    """
    Creates a cache and returns a LocalCache object. NOT THREAD-SAFE.
    """
    cache_config = os.path.join(local_path, CONFIG_FILENAME)
    if os.path.exists(cache_config):
        raise CacheError("A cache already exists in the specified location: " + local_path)

    try:
        os.makedirs(local_path)
    except OSError as e:
        pass #Directories exist

    ## Start creating config
    config = ConfigParser.SafeConfigParser()
    config.add_section(CONFIG_REPO_SECTION_NAME)

    config.set(CONFIG_REPO_SECTION_NAME, "remote_url", remote_url)
    config.set(CONFIG_REPO_SECTION_NAME, "config_version", CURRENT_REPO_CONFIG_VERSION)

    with open(cache_config, "w+") as f:
        config.write(f)

    return LocalCache(local_path)
 
def read_entry(file_location):
    if not os.path.exists(file_location):
        raise EntryError("Unable to find entry file under: " + file_location)

    #Get parser
    parser = ConfigParser.SafeConfigParser()
    try:#Read
        config = parser.read(file_location)
    except ConfigParser.ParsingError:
        raise EntryError("Error parsing the entry configuration file: " + file_location + " Please contact an administrator for help fixing this problem.")

    #Verify the config has been read 
    if config is None:
        raise EntryError("The config file: " + file_location + " appears to be empty. Please verify this is the location of a valid umpire cache entry.")

    #Verify the section we're looking for is there 
    if CONFIG_ENTRY_SECTION_NAME not in parser.sections():
        raise EntryError("The config file " + file_location + " appears to be missing the [umpire] section. Please verify this is the location of a valid umpire cache entry, or delete the containing folder and try again.")

    entry = EntryInfo()

    #Set file location
    entry.path = os.path.split(file_location)[0]

    #Attempt to parse the md5
    try:
        entry.md5 = parser.get(CONFIG_ENTRY_SECTION_NAME, "md5")
    except ConfigParser.Error:
        pass #TODO: Future
    if not entry.md5:
        pass #TODO: Future
    
    #Attempt to parse the entry product name
    try:
        entry.name = parser.get(CONFIG_ENTRY_SECTION_NAME, "name")
    except ConfigParser.Error:
        raise EntryError("Unable to parse name from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse name from entry file: " + file_location)
    
    #Attempt to parse the entry platform name
    try:
        entry.platform = parser.get(CONFIG_ENTRY_SECTION_NAME, "platform")
    except ConfigParser.Error:
        raise EntryError("Unable to parse platform from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse platform from entry file: " + file_location)

    #Attempt to parse the entry version 
    try:
        entry.version = parser.get(CONFIG_ENTRY_SECTION_NAME, "version")
    except ConfigParser.Error:
        raise EntryError("Unable to parse version from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse version from entry file: " + file_location)
    
    #Attempt to parse the entry config version 
    try:
        entry.config_version = parser.get(CONFIG_ENTRY_SECTION_NAME, "config_version")
    except ConfigParser.Error:
        raise EntryError("Unable to parse config_version from entry file: " + file_location)
    if not entry.config_version:
        raise EntryError("Unable to parse config_version from entry file: " + file_location)

    return entry

def write_entry(entry):

        ## Start creating config
        config = ConfigParser.SafeConfigParser()
        config_path = os.path.join(entry.path, CONFIG_FILENAME)
        config.add_section(CONFIG_ENTRY_SECTION_NAME)

        config.set(CONFIG_ENTRY_SECTION_NAME, "name", entry.name)
        config.set(CONFIG_ENTRY_SECTION_NAME, "platform", entry.platform)
        config.set(CONFIG_ENTRY_SECTION_NAME, "version", entry.version)
        config.set(CONFIG_ENTRY_SECTION_NAME, "config_version", CURRENT_ENTRY_CONFIG_VERSION)

        with open(os.path.join(entry.path, CONFIG_FILENAME), "w+") as f:
            return config.write(f)

class EntryInfo(object):
    config_version = None
    md5 = None
    name = None
    path = None
    platform = None
    version = None

    def lock(self):
        pass
    def unlock(self):
        pass

class EntryError(Exception):
    pass

class EntryLockError(Exception):
    pass

class CacheError(Exception):
    pass

class LocalCache(object):
    local_path = None
    settings_file = None
    remote_url = None
    config_version = None

    #Verifies local existance
    def __init__(self, cache_root):
        if not os.path.exists(cache_root):
            raise CacheError("The path '" + str(cache_root) + "' does not contain a valid umpire cache.")
        
        self.local_path = cache_root

        self.settings_file = os.path.join(cache_root,CONFIG_FILENAME)

        if not os.path.exists(self.settings_file):
            raise CacheError("The path '" + str(cache_root) + "' does not appear to be a valid umpire cache")
        config = None
        
        #Get parser
        parser = ConfigParser.SafeConfigParser()
        try:#Read
            config = parser.read(self.settings_file)
        except ConfigParser.ParsingError:
            raise CacheError("Error parsing the cache configuration file. Please contact an administrator for help fixing this problem.")

        #Verify the config has been read 
        if config is None:
            raise CacheError("The config file appears to be empty. Please verify this is the location of a valid umpire cache.")

        #Verify the section we're looking for is there 
        if CONFIG_REPO_SECTION_NAME not in parser.sections():
            raise CacheError("The config file appears to be missing the [umpire] section. Please verify this is the location of a valid umpire cache.")
        
        #Attempt to parse the data
        try:
            self.remote_url = parser.get(CONFIG_REPO_SECTION_NAME, "remote_url")
        except ConfigParser.Error:
            raise CacheError("Unable to determine the remote URL of this cache" + self.local_path)

        if not self.remote_url:
            raise CacheError("Unable to determine the remote URL of this cache:" + self.local_path)

        try:
            self.config_version = parser.get(CONFIG_REPO_SECTION_NAME, "config_version")
        except ConfigParser.Error:
            raise CacheError("Unable to determine the version of this cache: " + self.local_path)

        if not self.config_version:
            raise CacheError("Unable to determine the version of this cache: " + self.local_path)

    #Gets EntryInfo object, returns None if it doesn't exist.
    def get(self, platform, name, version):
        cs_entry_file = os.path.join(self.local_path, platform, name, version, ".umpire")
        entry_file = maestro.tools.path.get_case_insensitive_path(path=cs_entry_file)
        if entry_file is None:
            return None
        else:
            return read_entry(entry_file)
    
    #Locks a local entry
    def lock(self, entry):
        pass #TODO: Future

    def unlock(self, entry):
        pass #TODO: Future
        
    #Puts an archive of files into the cache 
    def put(self,archive_path, platform, name, version, unpack=True, force=False, keep_archive = False, keep_original = False):

        #Get Archive filename
        archive_filename = os.path.split(archive_path)[1]

        #Get path it would exist in the repo
        full_path = os.path.join(self.local_path, platform, name, version, archive_filename)
        entry_root = os.path.join(self.local_path, platform, name, version)
        case_insensitive_path = maestro.tools.path.get_case_insensitive_path(entry_root)

        #If file exists, and we're not force deleting it throw an error
        #TODO: modified for md5 additions, this will be obsolete
        if case_insensitive_path is not None and force is False:
            raise EntryError("Entry archive already exists under: " + case_insensitive_path + ". Please delete this entry, or enable force overwriting.")
        #Otherwise delete it
        elif case_insensitive_path is not None and force is True:
            os.remove(case_insensitive_path)

        try:
            os.makedirs(entry_root)
        except OSError as e:
            raise CacheError("Unknown error in creating cache entry: " + str(e))
        
        #Copy archive to cache location
        shutil.copy(archive_path, entry_root)

        #Remove original if we want
        if keep_original is False:
            os.remove(archive_path)

        unpacker = UnpackModule(None)
       
       #Unpack if necessary 
        if unpack is True:
            unpacker.destination_path = entry_root
            unpacker.file_path = full_path
            unpacker.delete_archive = not keep_archive
            unpacker.start()

        entry = EntryInfo()
        entry.name = name
        entry.version = version
        entry.platform = platform
        entry.path = entry_root
        #Write the entry
        write_entry(entry)
        #Lock the newly written entry
        self.lock(entry)
        if unpack is True:
            while unpacker.status != maestro.core.module.DONE:
                time.sleep(0.2)
                
            if unpacker.exception is not None:
                raise unpacker.exception

        
    #Set the repositories remote URL (updates local .umpire_repo file)
    def set_remote(url):
        pass #TODO: Future


#TODO: Write command line module
if __name__ == "__main__":
    try:
        cache = create_local_cache("./test", "s3://bucketname")
        cache.put("../../poco-1.4.6p2-win64.tar.gz", "win64","poco","1.4.6p2", keep_original = True)
        cache.get("win64","poco","1.4.6p2")
    finally:
        #pass
        shutil.rmtree("./test")



from setuptools import setup, find_packages
from distutils.core import setup

setup(name='Umpire',
      version='0.2.0a2',
      description='Generic dependency resolver.',
      author='Matthew Corner',
      author_email='mcorner@signiant.com',
      url='https://www.signiant.com',
      packages=find_packages(),
      license='MIT',
      install_requires=['MaestroOps>=0.3.0'],
      entry_points = {
          'console_scripts': [
              'umpire = umpire.umpire:entry'
              ]
          }
     )
